import colorlog
import logging
from concurrent_log_handler import ConcurrentRotatingFileHandler
import os
import inspect
from pathlib import Path
from functools import wraps
import datetime
from ..decorators.core import MetaSingleton


class Logger:
    """
    Examples:
        >>> logger = Logger(name='train-log', log_dir='./logs', print_debug=True)
        >>> logger.debug("hello", "list", [1, 2, 3, 4, 5])

        >>> logger2 = Logger.get_logger('train-log')
        >>> id(logger2) == id(logger)
        True

    """
    _saved_loggers = {}

    def __init__(self,
                 log_dir='./',
                 name='name',
                 debug_path="debug.log",
                 info_path='info.log',
                 warning_path='warn.log',
                 error_path='error.log',
                 print_debug=False,
                 print_info=False,
                 print_warning=False,
                 print_error=False,
                 single_mode=False,
                 level=logging.DEBUG,
                 tz='origin'
                 ):
        """
        Parameters
        ----------
            tz: time zone, to point china time zone you can use options: 'zh','ch','shanghai','beijing'。
        """

        self._colors_config = {
            'DEBUG': 'white',
            'INFO': 'cyan',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'bold_red',
        }

        if tz.lower() in ("zh", "ch", "shanghai", 'beijing'):
            logging.Formatter.converter = lambda sec, what: \
                (datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(hours=8)).timetuple()
        debug_path = Path(log_dir).joinpath(debug_path)
        info_path = Path(log_dir).joinpath(info_path)
        warning_path = Path(log_dir).joinpath(warning_path)
        error_path = Path(log_dir).joinpath(error_path)

        self._debug_logger = self._get_format_logger(f"debug-{name}", debug_path, level=logging.DEBUG,
                                                     stream=print_debug)
        self._info_logger = self._get_format_logger(f"info-{name}", info_path, level=logging.INFO, stream=print_info)
        self._warining_logger = self._get_format_logger(f"warning-{name}", warning_path, level=logging.WARNING,
                                                        stream=print_warning)
        self._error_logger = self._get_format_logger(f"error-{name}", error_path, level=logging.ERROR,
                                                     stream=print_error)
        self._single_mode = single_mode
        self._level = level
        self._param_dict = dict(
            log_dir=log_dir,
            debug_path=debug_path,
            info_path=info_path,
            warning_path=warning_path,
            error_path=error_path,
            print_debug=print_debug,
            print_info=print_info,
            print_warning=print_warning,
            print_error=print_error,
            single_mode=single_mode,
            level=level,
            tz=tz
        )
        self._saved_loggers[name] = self

    @classmethod
    def get_logger(cls, name):
        if name in cls._saved_loggers:
            return cls._saved_loggers[name]
        else:
            raise ValueError(
                f"`{name}` was not not find in saved_logges list, pls use `sparrow.log.Logger()` "
                f"to instantiate a logger object."
            )

    def debug(self, *msg, sep=' ', **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, "DEBUG", sep=sep)
        if self._level <= logging.DEBUG:
            self._debug_logger.debug(msg, **kwargs)

    def info(self, *msg, sep=' ', **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, "INFO", sep=sep)
        if self._level <= logging.INFO:
            self._info_logger.info(msg, **kwargs)
            if not self._single_mode:
                self._debug_logger.info(msg, **kwargs)

    def warning(self, *msg, sep=' ', **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, "WARNING", sep=sep)
        if self._level <= logging.WARNING:
            self._warining_logger.warning(msg, **kwargs)
            if not self._single_mode:
                self._debug_logger.warning(msg, **kwargs)
                self._info_logger.warning(msg, **kwargs)

    def error(self, *msg, sep=' ', **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, "ERROR", sep=sep)
        if self._level <= logging.ERROR:
            self._error_logger.error(msg, **kwargs)
            if not self._single_mode:
                self._debug_logger.error(msg, **kwargs)
                self._info_logger.error(msg, **kwargs)
                self._warining_logger.error(msg, **kwargs)

    @staticmethod
    def _get_format_msg(currentframe, msg: tuple, level, sep=" "):
        filename = os.path.basename(currentframe.f_back.f_code.co_filename)
        lineno = currentframe.f_back.f_lineno
        msg_list = [str(i) for i in msg]
        msg = sep.join(msg_list)
        msg = f"[{filename}]-[line:{lineno}]-{level} >>> " + msg
        return msg

    def _get_format_logger(self, name, log_abs_path, level=logging.INFO, stream=True):
        default_formats = {
            'color_format': '%(log_color)s%(asctime)s-%(message)s',
            'log_format': '%(asctime)s-%(message)s'
        }
        # default_formats = {
        #     'color_format': '%(log_color)s%(asctime)s-%(name)s-%(filename)s-[line:%(lineno)d]-%(levelname)s: %('
        #                     'message)s',
        #     'log_format': f'%(asctime)s-%(name)s-%(filename)s-[line:%(lineno)d]-%(levelname)s: %(message)s'
        # }

        log_path = Path(log_abs_path).absolute()
        log_dir = log_path.parent
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        # logger = logging.getLogger(name)
        logger = logging.Logger(name, level=level)

        stream_formatter = colorlog.ColoredFormatter(default_formats["color_format"],
                                                     log_colors=self._colors_config,
                                                     datefmt='%Y/%m/%d %H:%M:%S')

        file_formatter = logging.Formatter(default_formats["log_format"],
                                           datefmt='%Y-%m-%d %H:%M:%S')

        file_handler = ConcurrentRotatingFileHandler(filename=log_path,
                                                     maxBytes=10 * 1024 * 1024,
                                                     backupCount=10,
                                                     encoding='utf-8')
        file_handler.setFormatter(file_formatter)
        logger.addHandler(file_handler)

        if stream:
            stream_handler = logging.StreamHandler()
            stream_handler.setFormatter(stream_formatter)
            logger.addHandler(stream_handler)
        # logger.setLevel(level)
        return logger


# unused.
def findcaller(func):
    @wraps(func)
    def wrapper(*args):
        currentframe = inspect.currentframe()
        f = currentframe.f_back
        file_name = os.path.basename(f.f_code.co_filename)
        func_name = f.f_code.co_name
        line_num = f.f_lineno

        args = list(args)
        args.append(f'{os.path.basename(file_name)}.{func_name}.{line_num}')
        func(*args)

    return wrapper


class SingletonLogger(Logger, metaclass=MetaSingleton):
    def __init__(self,
                 log_dir='./',
                 name='name',
                 debug_path="debug.log",
                 info_path='info.log',
                 warning_path='warn.log',
                 error_path='error.log',
                 print_debug=False,
                 print_info=False,
                 print_warning=False,
                 print_error=False,
                 single_mode=False,
                 level=logging.DEBUG,
                 tz='origin'
                 ):
        super().__init__(
            log_dir=log_dir,
            name=name,
            debug_path=debug_path,
            info_path=info_path,
            warning_path=warning_path,
            error_path=error_path,
            print_debug=print_debug,
            print_info=print_info,
            print_warning=print_warning,
            print_error=print_error,
            single_mode=single_mode,
            level=level,
            tz=tz
        )

    @classmethod
    def getLogger(cls):
        """SingletonLogger is singleton，this is equivalent to using Logger () directly"""
        return cls()

    def copy(self):
        new_logger = object.__new__(Logger)
        new_logger.__init__(**self._param_dict)
        return new_logger



from .file_ops import yaml_load, yaml_dump, ppath, save, load, rm
from .decorator import benchmark
from .color.color_str import rgb_string
from .functions.core import clamp


_version_config = yaml_load(ppath("version-config.yaml", __file__))
__version__= _version_config['version']
print(f"{rgb_string(_version_config['name'], color='#34A853')} version: {rgb_string(__version__, color='#C9E8FF')}")



BLUE_E = "#1C758A"
BLUE_D = "#29ABCA"
BLUE_C = "#58C4DD"
BLUE_B = "#9CDCEB"
BLUE_A = "#C7E9F1"
TEAL_E = "#49A88F"
TEAL_D = "#55C1A7"
TEAL_C = "#5CD0B3"
TEAL_B = "#76DDC0"
TEAL_A = "#ACEAD7"
GREEN_E = "#699C52"
GREEN_D = "#77B05D"
GREEN_C = "#83C167"
GREEN_B = "#A6CF8C"
GREEN_A = "#C9E2AE"
YELLOW_E = "#E8C11C"
YELLOW_D = "#F4D345"
YELLOW_C = "#FFFF00"
YELLOW_B = "#FFEA94"
YELLOW_A = "#FFF1B6"
GOLD_E = "#C78D46"
GOLD_D = "#E1A158"
GOLD_C = "#F0AC5F"
GOLD_B = "#F9B775"
GOLD_A = "#F7C797"
RED_E = "#CF5044"
RED_D = "#E65A4C"
RED_C = "#FC6255"
RED_B = "#FF8080"
RED_A = "#F7A1A3"
MAROON_E = "#94424F"
MAROON_D = "#A24D61"
MAROON_C = "#C55F73"
MAROON_B = "#EC92AB"
MAROON_A = "#ECABC1"
PURPLE_E = "#644172"
PURPLE_D = "#715582"
PURPLE_C = "#9A72AC"
PURPLE_B = "#B189C6"
PURPLE_A = "#CAA3E8"
GREY_E = "#222222"
GREY_D = "#444444"
GREY_C = "#888888"
GREY_B = "#BBBBBB"
GREY_A = "#DDDDDD"
WHITE = "#FFFFFF"
BLACK = "#000000"
GREY_BROWN = "#736357"
DARK_BROWN = "#8B4513"
LIGHT_BROWN = "#CD853F"
PINK = "#D147BD"
LIGHT_PINK = "#DC75CD"
GREEN_SCREEN = "#00FF00"
ORANGE = "#FF862F"

# Abbreviated names for the "median" colors
BLUE = BLUE_C
TEAL = TEAL_C
GREEN = GREEN_C
YELLOW = YELLOW_C
GOLD = GOLD_C
RED = RED_C
MAROON = MAROON_C
PURPLE = PURPLE_C
GREY = GREY_C



from sparrow.version_ops import VersionControl

pkgname = "sparrow_tool"
pkgdir = "sparrow"
# vc = VersionControl(pkgname, pkgdir, version="0.1.6")
vc = VersionControl(pkgname, pkgdir, version=None)
vc.install()



from sparrow.version_ops import VersionControl
from git import Repo

pkgname = "sparrow_tool"
pkgdir = "sparrow"
# vc = VersionControl(pkgname, pkgdir, version="0.2.3")
vc = VersionControl(pkgname, pkgdir, version=None)
vc.update_version()
vc.update_readme()
# os.system("yapf -i -r ./sparrow")
# vc.upload_pypi()

repo = Repo('.')
repo.index.add(["README*.md", "requirements.txt", "workflow.py", 'setup.*', 'sparrow/version-config.yaml'])
repo.index.commit(f"[Upgrade] Version bump to [{vc.config['version']}]")
repo.create_tag(f"{vc.config['version']}")
remote = repo.remote()
remote.push(f"{vc.config['version']}")
remote.push()



import os
import sys
import shutil
from glob import glob
from .core import broadcast
import pickle


def _rm(path):
    """remove path
    """
    if os.path.exists(path):
        if os.path.isfile(path):
            os.remove(path)
        elif os.path.isdir(path):
            shutil.rmtree(path)
        else:
            print(f'{path} is illegal.')


@broadcast
def rm(PATH):
    """ Enhanced rm, support for regular expressions """
    path_list = glob(PATH)
    for path in path_list:
        _rm(path)


def path(string: str) -> str:
    """Adaptive to different platforms """
    platform = sys.platform.lower()
    if platform in ('linux', "darwin"):
        return string.replace('\\', '/')
    elif platform in ("win", "win32"):
        return string.replace('/', '\\')
    else:
        return string


def ppath(pathname, file=__file__) -> str:
    """Path in package"""
    return path(os.path.join(os.path.dirname(file), pathname))


def save(filename, file):
    with open(filename, 'wb') as fw:
        pickle.dump(file, fw)


def load(filename):
    with open(filename, 'rb') as fi:
        file = pickle.load(fi)
    return file


def yaml_dump(filepath, data):
    from yaml import dump
    try:
        from yaml import CDumper as Dumper
    except ImportError:
        from yaml import Dumper
    with open(filepath, "w", encoding='utf-8') as fw:
        fw.write(dump(data, Dumper=Dumper, allow_unicode=True, indent=4))


def yaml_load(filepath):
    from yaml import load
    try:
        from yaml import CLoader as Loader
    except ImportError:
        from yaml import Loader
    with open(filepath, 'r', encoding="utf-8") as stream:
        #     stream = stream.read()
        content = load(stream, Loader=Loader)
    return content



import keyboard
import time
from collections import deque
from sparrow.string.color_string import rgb_string, GREEN


def timer(dt=0.01):
    t0 = time.time()
    q = deque(maxlen=1)
    q.append(False)
    keyboard.add_hotkey('space', lambda: q.append(not q[0]))
    current_time = 0
    while True:
        time.sleep(dt)
        ct = time.time()
        if q[0]:
            t0 = ct
        else:
            current_time += ct - t0
            print(rgb_string(f"\r{current_time:.3f} secs", color=GREEN), end='')
            t0 = ct
        if keyboard.is_pressed('q'):
            break



from pathlib import Path
import inspect


def rel_to_abs(rel_path: str):
    """Return absolute path relative to the called file """
    currentframe = inspect.currentframe()
    f = currentframe.f_back
    current_path = Path(f.f_code.co_filename).parent
    return current_path / rel_path


if __name__ == "__main__":
    print(rel_to_abs("../emmm"))



# WIP

import importlib
from sys import modules
import pkgutil
import os

PRELOADED_MODULES = set()

result_pkg_list = []
def reload_pkg(module_path:str, top_pkg_name=None):
    global result_pkg_list
    print(module_path)
    for i in pkgutil.iter_modules([module_path]):
        result_pkg_list.append(i)
        if i.ispkg:
            subpkg_path = os.path.join(module_path, i.name)
#             subpkg_name =
            result_pkg_list.append(subpkg_path)
            reload_pkg(subpkg_path)


def reload():
    import importlib
    from sys import modules
    for module in set(modules.values()) - PRELOADED_MODULES:
        try:
            importlib.reload(module)
        except:
            # there are some problems that are swept under the rug here
            pass



import os

from .file_ops import yaml_load, yaml_dump, rm
from .string_ops import string_add


class VersionControl:
    def __init__(self,
                 pkgname,
                 pkgdir,
                 version=None,
                 filename="version-config.yaml",
                 ):

        self.config = None
        self._pkgname = pkgname
        self._config_path = os.path.join(pkgdir, filename)
        if version is None:
            try:
                self.get_config()
            except:
                print(f"{filename} was not exist, created now.")
                self.gen_config('0.0.0')
        else:
            self.gen_config(version)

    def gen_config(self, version="0.0.0"):
        config = {"name": self._pkgname, "version": version}
        self.config = config
        yaml_dump(self._config_path, config)

    def get_config(self):
        config = yaml_load(self._config_path)
        self.config = config
        return config

    def set_version(self, version):
        self.config['version'] = version

    def save_config(self):
        yaml_dump(self._config_path, self.config)

    def update_version(self, version_step=1):
        self.config['version'] = string_add(self.config["version"], version_step)
        yaml_dump(self._config_path, self.config)

    def clean_config_file(self):
        os.remove(self._config_path)

    def update_readme(self, readme_path="README.md",
                      requirements_path="requirements.txt",
                      license="GNU_GPL--v3",
                      author="kunyuan",
                      replace_flag=19 * '-'):
        with open(readme_path, 'r', encoding='UTF-8') as fr:
            readme = fr.read()
        with open(requirements_path, 'r', encoding="utf-8") as fr:
            requirements = fr.read().strip("\n")
        replace_begin = f"""# {self._pkgname}
[![image](https://img.shields.io/badge/Pypi-{self.config['version']}-green.svg)](https://pypi.org/project/{self._pkgname})
[![image](https://img.shields.io/badge/python-3.6+-blue.svg)](https://www.python.org/)
[![image](https://img.shields.io/badge/license-{license}-blue.svg)](LICENSE)
[![image](https://img.shields.io/badge/author-{author}-orange.svg?style=flat-square&logo=appveyor)](https://github.com/beidongjiedeguang)

## Requirements
```python
{requirements}
```

"""
        readme_list = readme.split(replace_flag)
        readme_list[0] = replace_begin
        new_readme = replace_flag.join(readme_list)

        with open(readme_path, 'w', encoding='UTF-8') as fo:
            fo.write(new_readme)

    def upload_pypi(self):
        pkgname = self._pkgname
        rm('build', 'dist', 'eggs', f'{pkgname}.egg-info')
        os.system('python setup.py sdist bdist_wheel')
        os.system('twine upload dist/*')
        rm('build', 'dist', 'eggs', f'{pkgname}.egg-info')

    def install(self):
        pkgname = self._pkgname
        rm('build', 'dist', 'eggs', f'{pkgname}.egg-info')
        os.system(f'pip uninstall {pkgname} -y && python setup.py install')
        rm('build', 'dist', 'eggs', f'{pkgname}.egg-info')
